#!/bin/python
import pymongo
from pymongo.dbref import DBRef
from minimongo import config
from pymongo.cursor import Cursor as PyMongoCursor
from pymongo.objectid import ObjectId

class MongoCollection(object):
    """Container class for connection to db & mongo collection settings."""
    def __init__(self,
                 host=None, port=None, database=None, collection=None):
        if not host:
            host = config.MONGODB_HOST
        if not port:
            port = config.MONGODB_PORT
        self.host = host
        self.port = port
        self.database = database
        self.collection = collection


def cursor_wrapped(wrapped):
    def method(cursor, *args, **kwargs):
        return Cursor(results=wrapped(cursor._results, *args, **kwargs),
                      obj_type=cursor._obj_type)
    cursor_wrapped.__doc__ = wrapped.__doc__
    return method


class Cursor(object):
    """Simple wrapper around the cursor (iterator) that comes back from
    pymongo.  We do this so that when you iterate through results from a
    find, you get a generator of Model objects, not a bunch of dicts. """
    def __init__(self, results, obj_type):
        self._obj_type = obj_type
        self._results = results

    def count(self):
        """Analog of the normal count() operation on MongoDB cursors.
        Returns the number of items matching this query."""
        return self._results.count()

    # rewind = cursor_wrapped(PyMongoCursor.rewind)
    clone = cursor_wrapped(PyMongoCursor.clone)
    limit = cursor_wrapped(PyMongoCursor.limit)
    skip = cursor_wrapped(PyMongoCursor.skip)
    sort = cursor_wrapped(PyMongoCursor.sort)

    def __iter__(self):
        for i in self._results:
            yield(self._obj_type(i))


class Meta(type):
    """Metaclass for our model class.  Inspects the class variables, looks
    for 'mongo' and uses that to connect to the database. """

    # A very rudimentary connection pool:
    _connections = {}

    def __new__(mcs, name, bases, data):
        # Pull fields out of the MongoCollection object to get the database
        # connection parameters, etc.
        collection_info = data['mongo']
        if 'indices' in data:
            index_info = data['indices']
        else:
            index_info = []
        host = collection_info.host
        port = collection_info.port
        dbname = collection_info.database
        collname = collection_info.collection

        new_cls = super(Meta, mcs).__new__(mcs, name, bases, data)

        # This constructor runs on the Model class as well as the derived
        # classes.  When we're a Model, we don't have a proper
        # configuration, so we just skip the connection stuff below.
        if name == 'Model':
            new_cls.db = None
            new_cls.collection = None
            return new_cls

        if not (host and port and dbname and collname):
            raise Exception(
                'minimongo Model %s %s improperly configured: %s %s %s %s' % (
                    mcs, name, host, port, dbname, collname))

        hostport = (host, port)
            # Check the connection pool for an existing connection.
        if hostport in mcs._connections:
            connection = mcs._connections[hostport]
        else:
            connection = pymongo.Connection(host, port)
        mcs._connections[hostport] = connection
        new_cls.db = connection[dbname]
        new_cls.collection = new_cls.db[collname]
        new_cls._collection_name = collname
        new_cls._database_name = dbname
        new_cls._index_info = index_info

        # Generate all our indices automatically when the class is
        # instantiated.  This will result in calls to pymongo's
        # ensure_index() method at import time, so import all your models up
        # front.
        new_cls.auto_index()

        return new_cls

    def auto_index(mcs):
        """Build all indices for this collection specified in the definition
        of the Model."""
        for index in mcs._index_info:
            index.ensure(mcs.collection)

    def collection_name(mcs):
        """Return the name of the MongDB collection for the current Model."""
        return mcs._collection_name

    def database_name(mcs):
        """Return the name of the MongDB database for the current Model."""
        return mcs._database_name

    def from_dbref(mcs, dbref):
        """Given a DBRef, return an instance."""
        return mcs.find_one({'_id': dbref.id})

    def find(mcs, *args, **kwargs):
        """Passthrough to pymongo's find() method, and wrap the results in a
        Cursor object so that we get Model objects while iterating."""
        results = mcs.collection.find(*args, **kwargs)
        return Cursor(results, mcs)

    def find_one(mcs, *args, **kwargs):
        """Passthrough to pymongo's find_one() method, and wrap the results
        in a Model object of the correct type."""
        data = mcs.collection.find_one(*args, **kwargs)
        if data:
            return mcs(data)
        return None

    def __getattribute__(mcs, *args):
        """This gets invoked for things that look like classmethods.  First
        we try the attribute from self, then from the collection, then from
        the db."""
        try:
            ret = object.__getattribute__(mcs, *args)
        except AttributeError:
            try:
                ret = object.__getattribute__(mcs.collection, *args)
            except AttributeError:
                ret = object.__getattribute__(mcs.db, *args)
        return ret


class Model(object):
    """Base class for all Minimongo objects.  Derive from this class."""
    __metaclass__ = Meta
    mongo = MongoCollection(host=None, port=None,
                            database=None, collection=None)

    def __init__(self, data=None):
        if data:
            self.__dict__ = data
        else:
            self.__dict__ = {}

    def dbref(self):
        """Return an instance of a DBRef for the current object."""
        if not self._id:
            self._id = ObjectId()
        assert self._id != None, "ObjectId must be valid to create DBRef"
        return DBRef(collection=self.collection_name,
                     id=self._id,
                     database=self.database_name)

    @property
    def id(self):
        """Return the MongoDB _id value."""
        return self._id

    @property
    def rawdata(self):
        """Return the raw document data as a dict."""
        return self.__dict__

    @property
    def database_name(self):
        """Return the name of the MongoDB Database for this class."""
        return type(self)._database_name

    @property
    def collection_name(self):
        """Return the name of the MongoDB collection for this class."""
        return type(self)._collection_name

    def remove(self):
        """Delete this object."""
        return self.collection.remove(self._id)

    def update(self):
        """Update (write) this object."""
        self.collection.update(
            {'_id': self._id},
            self.__dict__)
        return self

    def save(self):
        """Save this Model to it's mongo collection"""
        self.collection.save(self.__dict__)
        return self

    def __str__(self):
        ret = type(self).__name__ + '(' + str(self.__dict__) + ')'
        return ret

    def __unicode__(self):
        ret = type(self).__name__ + u'(' + str(self.__dict__) + u')'
        return ret


class Index(object):
    """Just a simple container class holding the arguments that are passed
    directly to pymongo's ensure_index method."""
    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    def ensure(self, collection):
        """Call pymongo's ensure_index on the given collection with the
        stored args."""
        return collection.ensure_index(*self._args, **self._kwargs)





import unittest

from minimongo.model import Model, MongoCollection, Index


class TestModel(Model):
    """Model class for test cases."""
    mongo = MongoCollection(database='test', collection='minimongo_test')
    indices = (Index('x'),)


class TestModelUnique(Model):
    mongo = MongoCollection(database='test', collection='minimongo_unique')
    indices = (Index('x', unique=True),)


def assertContains(iterator, instance):
    """Given an iterable of Models, make sure that the instance (of a Model)
    is inside the iterable."""
    for i in iterator:
        if i.rawdata == instance.rawdata:
            return True
    return False


class TestSimpleModel(unittest.TestCase):
    """Main test case."""
    def setUp(self):
        """unittest setup, drop the whole collection, then rebuild indices
        before starting each test."""
        TestModel.drop()
        TestModel.auto_index()

        TestModelUnique.drop()
        TestModelUnique.auto_index()

    def test_creation(self):
        """Test simple object creation and querying via find_one."""
        dummy_m = TestModel({'x': 1, 'y': 1})
        dummy_m.save()
        dummy_n = TestModel.find_one({'x': 1})
        # Make sure that the find_one method returns the right type.
        self.assertEqual(type(dummy_n), TestModel)
        # Make sure that the contents are the same.
        self.assertEqual(dummy_n.rawdata, dummy_m.rawdata)

        # Make sure that our internal representation is what we expect (and
        # no extra fields, etc.)
        self.assertEqual(dummy_m.rawdata, {'x': 1, 'y': 1, '_id': dummy_m._id})
        self.assertEqual(dummy_n.rawdata, {'x': 1, 'y': 1, '_id': dummy_n._id})

    def test_index_existance(self):
        """Test that indexes were created properly."""
        indices = TestModel.index_information()
        self.assertEqual(indices['x_1'],
                         {'key': [('x', 1)]})


    def test_unique_index(self):
        """Test behavior of indices with unique=True"""
        # This will work (y is undefined)
        x1_a = TestModelUnique({'x': 1}).save()
        x1_b = TestModelUnique({'x': 1}).save()
        # Assert that there's only one object in the collection, even though
        # we inserted two.  The uniqueness constraint on the index has
        # dropped one of the inserts (silently, I guess).x
        self.assertEqual(TestModelUnique.find().count(), 1)

        # Even if we use different values for y, it's still only one object:
        x2_a = TestModelUnique({'x': 2, 'y': 1}).save()
        x2_b = TestModelUnique({'x': 2, 'y': 2}).save()
        # There are now 2 objects, one with x=1, one with x=2.
        self.assertEqual(TestModelUnique.find().count(), 2)

    def test_queries(self):
        """Test some more complex query forms."""
        dummy_a = TestModel({'x': 1, 'y': 1}).save()
        dummy_b = TestModel({'x': 1, 'y': 2}).save()
        dummy_c = TestModel({'x': 2, 'y': 2}).save()
        dummy_d = TestModel({'x': 2, 'y': 1}).save()

        found_x1 = TestModel.find({'x': 1})
        found_y1 = TestModel.find({'y': 1})
        found_x2y2 = TestModel.find({'x': 2, 'y': 2})

        list_x1 = list(found_x1)
        list_y1 = list(found_y1)
        list_x2y2 = list(found_x2y2)

        self.assertEqual(type(list_x1[0]), TestModel)

        self.assertTrue(assertContains(list_x1, dummy_a))
        self.assertTrue(assertContains(list_x1, dummy_b))
        self.assertTrue(assertContains(list_y1, dummy_a))
        self.assertTrue(assertContains(list_y1, dummy_d))
        self.assertEqual(list_x2y2[0].rawdata, dummy_c.rawdata)

    def test_deletion(self):
        """Test deleting an object from a collection."""
        dummy_m = TestModel()
        dummy_m.x = 100
        dummy_m.y = 200
        dummy_m.save()

        dummy_n = TestModel.find({'x': 100})
        self.assertEqual(dummy_n.count(), 1)
        for i in dummy_n:
            i.remove()

        dummy_m = TestModel.find({'x': 100})
        self.assertEqual(dummy_m.count(), 0)

    def test_complex_types(self):
        """Test lists as types."""
        dummy_m = TestModel()
        dummy_m.l = ['a', 'b', 'c']
        dummy_m.x = 1
        dummy_m.save()

        dummy_n = TestModel.find_one({'x': 1})

        ml = dummy_m.l
        nl = dummy_n.l
        self.assertEqual(ml, nl)

    def test_delete_field(self):
        """Test deleting a single field from an object."""
        dummy_m = TestModel({'x': 1, 'y': 2})
        dummy_m.save()
        del(dummy_m.x)
        dummy_m.save()

        n = TestModel.find_one({'y': 2})
        self.assertEqual(n.rawdata, {'y': 2, '_id': dummy_m._id})

    def test_count_and_fetch(self):
        """Test counting methods on Cursors. """
        dummy_d = TestModel({'x': 1, 'y': 4}).save()
        dummy_b = TestModel({'x': 1, 'y': 2}).save()
        dummy_a = TestModel({'x': 1, 'y': 1}).save()
        dummy_c = TestModel({'x': 1, 'y': 3}).save()

        find_x1 = TestModel.find({'x': 1}).sort('y')
        self.assertEqual(find_x1.count(), 4)
        list_x1 = list(find_x1)
        self.assertEqual(list_x1[0].rawdata, dummy_a.rawdata)
        self.assertEqual(list_x1[1].rawdata, dummy_b.rawdata)
        self.assertEqual(list_x1[2].rawdata, dummy_c.rawdata)
        self.assertEqual(list_x1[3].rawdata, dummy_d.rawdata)

    def test_fetch_and_limit(self):
        """Test counting methods on Cursors. """
        dummy_d = TestModel({'x': 1, 'y': 4}).save()
        dummy_b = TestModel({'x': 1, 'y': 2}).save()
        dummy_a = TestModel({'x': 1, 'y': 1}).save()
        dummy_c = TestModel({'x': 1, 'y': 3}).save()

        find_x1 = TestModel.find({'x': 1}).limit(2).sort('y')
        # Huh, calling count() on a find() with a limit() returns the total
        # number of elements matched.  Maybe I'm calling it wrong?
        # self.assertEqual(find_x1.count(), 2)
        list_x1 = list(find_x1)
        self.assertEqual(len(list_x1), 2)
        self.assertEqual(list_x1[0].rawdata, dummy_a.rawdata)
        self.assertEqual(list_x1[1].rawdata, dummy_b.rawdata)

    def test_dbref(self):
        """Test generation of DBRef objects, and querying via DBRef
        objects."""
        dummy_a = TestModel({'x': 1, 'y': 999}).save()
        ref_a = dummy_a.dbref()

        dummy_b = TestModel.from_dbref(ref_a)
        self.assertEqual(dummy_a.rawdata, dummy_b.rawdata)

    def test_db_and_collection_names(self):
        """Test the methods that return the current class's DB and
        Collection names."""
        dummy_a = TestModel({'x': 1})
        self.assertEqual(dummy_a.database_name, 'test')
        self.assertEqual(TestModel.database_name(), 'test')
        self.assertEqual(dummy_a.collection_name, 'minimongo_test')
        self.assertEqual(TestModel.collection_name(), 'minimongo_test')

if __name__ == '__main__':
    unittest.main()



#!/bin/python
"""
Provides the 2 symbols MONGODB_HOST and MONGODB_PORT that are used my
minimingo to connect to the proper database.

Uses the following 3 strategies:
1. import os.environ['MINIMONGO_SETTINGS_MODULE']
2. import 'minimongo.app_config'
3. default values (localhost, 27017)
"""
import sys
import os

# Default values for MONGODB_HOST and MONGODB_PORT if no custom config
# module is specified, or if we're unable to 'from minimongo.app_config import
# MONGODB_HOST, MONGODB_PORT'
MONGODB_HOST = 'localhost'
MONGODB_PORT = 27017


# Taken from Django, which says "Taken from Python 2.7..."
def _resolve_name(name, package, level):
    """Return the absolute name of the module to be imported."""
    if not hasattr(package, 'rindex'):
        raise ValueError("'package' not set to a string")
    dot = len(package)
    for x in xrange(level, 1, -1):
        try:
            dot = package.rindex('.', 0, dot)
        except ValueError:
            raise ValueError("attempted relative import beyond top-level "
                              "package")
    return "%s.%s" % (package[:dot], name)


def import_module(name, package=None):
    """Import a module.

    The 'package' argument is required when performing a relative import. It
    specifies the package to use as the anchor point from which to resolve the
    relative import to an absolute import.

    """
    if name.startswith('.'):
        if not package:
            raise TypeError("relative imports require the 'package' argument")
        level = 0
        for character in name:
            if character != '.':
                break
            level += 1
        name = _resolve_name(name[level:], package, level)
    __import__(name)
    return sys.modules[name]

if __name__ != '__main__':
    try:
        settings_module_name = os.environ['MINIMONGO_SETTINGS_MODULE']
    except KeyError, e:
        try:
            settings_module_name = 'minimongo.app_config'
            module = import_module(settings_module_name)
            MONGODB_HOST = module.MONGODB_HOST
            MONGODB_PORT = module.MONGODB_PORT
        except ImportError, e:
            # Can't import either MINIMONGO_SETTINGS_MODULE, or
            # minimongo_config, so we just return, knowing that the default
            # values will be used.
            pass
